﻿using Kivii.Finances.Entities;
using Kivii.Linq;
using Kivii.Linq.Dapper;
using Kivii.Web;
using System;
using System.Collections.Generic;
using System.Data;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace Kivii.Finances.Transforms
{

    [Api(Description = "插入发票")]
    [RequiresAnyRole(MemberRoles.Everyone)]
    public class InvoiceAccept : RestfulExecution<Invoice>
    {
        public Invoice Item { get; set; }
        public List<Invoice> Items { get; set; }

        public List<Payment> Payments { get; set; }

        // 定义批处理大小常量
        private const int BATCH_SIZE = 100;
        // 定义处理超时时间
        private readonly TimeSpan PROCESSING_TIMEOUT = TimeSpan.FromMinutes(2);

        public override object OnExecution(IRequest req, IResponse res)
        {
            (Item == null && Items.IsNullOrEmpty()).ThrowIfTrue("开票结果内容不能为空！");
            if (Items.IsNullOrEmpty()) Items = new List<Invoice>();
            if (Item != null) Items.Add(Item);
            
            // 性能优化：首先检查数据量，提前预警
            if (Items.Count > BATCH_SIZE)
            {
                KiviiContext.Logger.Warn($"正在处理大批量发票数据: {Items.Count}条，可能需要较长时间");
            }
            
            foreach (var item in Items)
            {
                (item.OperateTime == DateTime.MinValue || item.OperateTime == DateTime.MaxValue).ThrowIfTrue("缺少开票日期信息!");
                //item.SerialCode.ThrowIfNullOrEmpty("缺少发票代码信息!");
                item.SerialNumber.ThrowIfNullOrEmpty("缺少发票号码信息!");
                item.Type.ThrowIfNullOrEmpty("缺少发票类型信息!");
                (item.TaxRate < 0 || item.TaxRate > 1).ThrowIfTrue("缺少发票税率信息!");
                (item.Amount == 0).ThrowIfTrue("缺少发票金额信息!");
                item.Details.ThrowIfNullOrEmpty("缺少发票明细信息!");
                (item.Amount != item.Details.Sum(o => o.Amount)).ThrowIfTrue("明细金额和传入发票金额不一致!");
            }
            if (!Payments.IsNullOrEmpty())
            {
                (Payments.Exists(o => o.Kvid == Guid.Empty)).ThrowIfTrue("不存在的到账信息!");
                (Payments.Exists(o => o.Amount <= 0)).ThrowIfTrue("未指定到账开票金额!");
                (Payments.Sum(o => o.Amount) != Items.Sum(o => o.Amount)).ThrowIfTrue("到账开票金额和发票金额不一致!");
            }
            
            #region 启用缓存,将当前人创建批次数据存入,处理完毕或者接口报错后清除,确保不会重复创建
            var cache = KiviiContext.GetCacheClient();
            var cacheKey = KiviiContext.GetUrnKey($"{this.GetType().FullName}_Request_{KiviiContext.CurrentMember.FullName}_{KiviiContext.CurrentMember.Kvid}");
            var cacheValue = cache.Get<int?>(cacheKey);
            if (cacheValue != null)
            {
                if (cacheValue < 0) throw new Exception($"仍在处理中,请稍后查看结果!");
                else throw new Exception($"当前正在处理中,剩余{cacheValue}条,请稍后查看结果!");
            }
            int? caches = Items.Count;
            //caches.AddRange(Items.ConvertAll(o => o.SerialNumber));
            cache.Set(cacheKey, caches, TimeSpan.FromMinutes(5));
            #endregion

            // 数据库连接初始化（移到外层以确保异常时能正确关闭）
            var conn = KiviiContext.GetOpenedDbConnection<Invoice>();
            var trans = null as IDbTransaction;
            var lockAcquired = false;
            var lockName = $"invoice_lock_{KiviiContext.CurrentMember.Kvid}";
            
            try
            {
                // 优化2：使用数据库锁防止并发问题
                try
                {
                    // 使用数据库锁防止同一用户并发提交
                    var lockResult = conn.ExecuteScalar<int>("SELECT GET_LOCK(@lockName, 10)", 
                        new { lockName = lockName });
                    
                    if (lockResult != 1)
                    {
                        throw new Exception("系统繁忙，请稍后再试");
                    }
                    lockAcquired = true;
                }
                catch (Exception lockEx)
                {
                    // 如果数据库不支持GET_LOCK，降级到缓存锁
                    KiviiContext.Logger.Warn($"数据库锁获取失败，降级使用缓存锁: {lockEx.Message}");
                }
                
                //var exists = conn.Select<Invoice>(o => Sql.In(o.SerialNumber, Items.ConvertAll(p => p.SerialNumber)));
                //(conn.Exists<Invoice>(o => Sql.In(o.SerialNumber, Items.ConvertAll(p => p.SerialNumber)))).ThrowIfTrue("存在重复录入的发票信息");

                List<Payment> preCorrelatingPayments = new List<Payment>();
                List<Payment> parentPayments = new List<Payment>();//也要更新父级的AmountInvoice
                var insertInvoices = new List<Invoice>();
                var insertInvoiceDetails = new List<InvoiceDetail>();

                if (!Payments.IsNullOrEmpty())
                {
                    var existPayments = conn.SelectByIds<Payment>(Payments.ConvertAll(o => o.Kvid));
                    existPayments.ThrowIfNullOrEmpty("未找到指定的到账信息!");
                    (existPayments.Count != Payments.Count).ThrowIfTrue("指定的到账信息与查询到的到账信息不匹配!");
                    existPayments.Exists(o => o.Type != PaymentType.Bank && o.Type != PaymentType.AliPay && o.Type != PaymentType.Cash && o.Type != PaymentType.Pos && o.Type != PaymentType.Split && o.Type != PaymentType.WeChat).ThrowIfTrue("存在不正确的到账类型");
                    foreach (var item in existPayments)
                    {
                        var payment = Payments.FirstOrDefault(o => o.Kvid == item.Kvid);
                        if (payment == null) continue;
                        if (item.Amount - item.AmountInvoice < payment.Amount) throw new Exception($"所选到账{item.SerialNumber},不足以继续开票指定金额{payment.Amount}元,当前到账剩余开票金额为:{item.Amount - item.AmountInvoice}元");
                        if (item.Amount - item.AmountInvoice >= payment.Amount)//说明选中的到账足以开票指定的金额
                        {
                            //item.Amount = payment.Amount;
                            //item.AmountInvoice = 0;
                            item.RemoveAllOnlyProperties();
                            preCorrelatingPayments.Add(item);
                        }
                    }
                    (preCorrelatingPayments.Sum(o => o.Amount - o.AmountInvoice) != Items.Sum(o => o.Amount)).ThrowIfTrue("指定的到账开票金额和发票金额不一致!");
                }

                var category = preCorrelatingPayments.IsNullOrEmpty() ? InvoiceApplyType.Debit : InvoiceApplyType.Payment;
                
                // 创建任务以支持超时处理
                var processingTask = Task.Run(() => {
                    foreach (var info in Items)
                    {
                        //重复的发票,如果传了对应到账,那重复的发票就要抛异常,否则跳过
                        if (conn.Exists<Invoice>(o => o.SerialNumber == info.SerialNumber))
                        {
                            if (!Payments.IsNullOrEmpty()) throw new Exception($"存在重复录入的发票信息:{info.SerialNumber}");
                            else continue;
                        }
                        #region 开票成功的情况处理
                        var invoiceKvid = Guid.NewGuid();
                        var invoice = new Invoice();
                        invoice.RootKvid = invoiceKvid;
                        invoice.Kvid = invoiceKvid;
                        //invoice.ApplyKvid = ApplyKvid;
                        invoice.OwnerKvid = info.OwnerKvid;
                        invoice.OwnerName = info.OwnerName;
                        if (invoice.OwnerKvid == Guid.Empty && invoice.OwnerName.IsNullOrEmpty())
                        {
                            invoice.OwnerKvid = KiviiContext.CurrentMember.DepartmentKvid;
                            invoice.OwnerName = KiviiContext.CurrentMember.DepartmentName;
                        }
                        invoice.Currency = CurrencyUnit.CNY;
                        invoice.Type = info.Type;
                        invoice.Category = info.Category.IsNullOrEmpty() ? category.ToString() : info.Category;
                        invoice.OperateTime = info.OperateTime;
                        invoice.SerialCode = info.SerialCode;
                        invoice.SerialNumber = info.SerialNumber;
                        //invoice.Status = 1;
                        invoice.TaxRate = info.TaxRate;
                        invoice.Amount = info.Amount;
                        invoice.AmountUntaxed = Math.Round(invoice.Amount / (1 + invoice.TaxRate), 2);
                        invoice.AmountTax = invoice.Amount - invoice.AmountUntaxed;
                        invoice.PayeeKvid = info.PayeeKvid;
                        invoice.PayeeName = info.PayeeName;
                        invoice.PayeeTaxNumber = info.PayeeTaxNumber;
                        invoice.PayeeCompanyAddress = info.PayeeCompanyAddress;
                        invoice.PayeePhone = info.PayeePhone;
                        invoice.PayeeRegisteredBank = info.PayeeRegisteredBank;
                        invoice.PayeeBankAccount = info.PayeeBankAccount;
                        invoice.PayerKvid = info.PayerKvid;
                        invoice.PayerName = info.PayerName;
                        invoice.PayerTaxNumber = info.PayerTaxNumber;
                        invoice.PayerCompanyAddress = info.PayerCompanyAddress;
                        invoice.PayerPhone = info.PayerPhone;
                        invoice.PayerRegisteredBank = info.PayerRegisteredBank;
                        invoice.PayerBankAccount = info.PayerBankAccount;
                        invoice.OperatorName = info.OperatorName;
                        invoice.Remark = info.Remark;

                        invoice.Metadata = new Dictionary<string, string>();
                        if (!info.Metadata.IsNullOrEmpty())
                        {
                            invoice.Metadata = info.Metadata;
                            invoice.Metadata["PayeeOperatorName"] = info.Metadata.ContainsKey("PayeeOperatorName") ? info.Metadata["PayeeOperatorName"] : "";
                            invoice.Metadata["ReviewerName"] = info.Metadata.ContainsKey("ReviewerName") ? info.Metadata["ReviewerName"] : "";
                        }
                        insertInvoices.Add(invoice);

                        foreach (var infoDetail in info.Details)
                        {
                            var detail = new InvoiceDetail();
                            detail.PopulateInstance(infoDetail);
                            detail.InvoiceKvid = invoiceKvid;
                            detail.TaxRate = invoice.TaxRate;
                            detail.AmountUntaxed = invoice.AmountUntaxed;
                            detail.Amount = invoice.Amount;
                            detail.AmountTax = invoice.AmountTax;
                            detail.Quantity = detail.Quantity <= 0 ? 1 : detail.Quantity;
                            detail.QuantityUnitPrice = Math.Round(detail.Amount / detail.Quantity, 2);
                            detail.QuantityUnitPriceUntaxed = Math.Round(detail.AmountUntaxed / detail.Quantity, 2);
                            insertInvoiceDetails.Add(detail);
                        }
                        #endregion
                    }
                    
                    return true;
                });
                
                // 添加超时处理
                var timeoutTask = Task.Delay(PROCESSING_TIMEOUT);
                if (Task.WhenAny(processingTask, timeoutTask).Result == timeoutTask)
                {
                    throw new TimeoutException($"处理发票数据超时，请减少批量处理数量（当前{Items.Count}条）");
                }
                
                // 检查处理任务是否成功
                if (!processingTask.Result)
                {
                    throw new Exception("处理发票数据失败");
                }

                if (!preCorrelatingPayments.IsNullOrEmpty())
                {
                    var correlatedInvoices = insertInvoices.Correlating(preCorrelatingPayments);
                    insertInvoices.AddRange(correlatedInvoices);
                    var currentParentKvids = preCorrelatingPayments.Where(o => o.Type == PaymentType.Split).ToList().ConvertAll(p => p.ParentKvid);
                    if (!currentParentKvids.IsNullOrEmpty()) parentPayments = conn.Select<Payment>(o => o.OffsetKvid == Guid.Empty && Sql.In(o.Kvid, currentParentKvids));
                }
                var rtns = new RestfulUpdateResponse<Invoice>();
                rtns.Results = new List<Invoice>();
                
                // 开启事务
                trans = conn.OpenTransaction(System.Data.IsolationLevel.ReadCommitted);
                
                try
                {
                    // 性能优化：批量插入发票
                    if (!insertInvoices.IsNullOrEmpty())
                    {
                        // 分批处理数据，避免单次处理过多
                        for (int i = 0; i < insertInvoices.Count; i += BATCH_SIZE)
                        {
                            var batchInvoices = insertInvoices.Skip(i).Take(BATCH_SIZE).ToList();
                            
                            // 处理批次中的每个发票
                            foreach (var item in batchInvoices)
                            {
                                if (item.Type == "Relation") 
                                    item.SerialNumber = item.GetSubSerialNumber(conn);
                                
                                conn.Insert(item);
                                if (item.Type != "Relation") 
                                    rtns.Results.Add(item);
                                
                                // 更新缓存计数
                                caches--;
                                try {
                                    cache.Replace(cacheKey, caches);
                                }
                                catch (Exception cacheEx) {
                                    // 缓存更新失败不影响主流程
                                    KiviiContext.Logger.Warn($"缓存更新失败: {cacheEx.Message}");
                                }
                            }
                        }
                    }
                    
                    // 性能优化：批量插入发票明细
                    if (!insertInvoiceDetails.IsNullOrEmpty())
                    {
                        // 如果数据库支持批量插入，使用批量插入
                        try {
                            conn.InsertAll(insertInvoiceDetails);
                        }
                        catch {
                            // 降级到循环插入
                            for (int i = 0; i < insertInvoiceDetails.Count; i += BATCH_SIZE)
                            {
                                var batch = insertInvoiceDetails.Skip(i).Take(BATCH_SIZE).ToList();
                                foreach (var detail in batch)
                                {
                                    conn.Insert(detail);
                                }
                            }
                        }
                    }
                    
                    // 更新关联的付款记录
                    if (!preCorrelatingPayments.IsNullOrEmpty())
                    {
                        for (int i = 0; i < preCorrelatingPayments.Count; i += BATCH_SIZE)
                        {
                            var batch = preCorrelatingPayments.Skip(i).Take(BATCH_SIZE).ToList();
                            foreach (var payment in batch)
                            {
                                conn.UpdateOnly(payment);
                            }
                        }
                    }
                    
                    // 更新父级的付款记录
                    if (!parentPayments.IsNullOrEmpty())
                    {
                        foreach (var payment in parentPayments)
                        {
                            payment.RecalculateAmountInvoice(conn);
                        }
                    }
                    
                    trans.Commit();
                    
                    // 清理缓存
                    try {
                        cache.Remove(cacheKey);
                    }
                    catch (Exception cacheEx) {
                        // 缓存清理失败仅记录警告
                        KiviiContext.Logger.Warn($"缓存清理失败: {cacheEx.Message}");
                    }
                    
                    return rtns;
                }
                catch (Exception ex)
                {
                    // 事务回滚
                    if (trans != null)
                    {
                        try { 
                            trans.Rollback(); 
                        } 
                        catch (Exception rollbackEx) {
                            KiviiContext.Logger.Error($"事务回滚失败: {rollbackEx.Message}");
                        }
                    }
                    
                    // 清理缓存
                    try {
                        cache.Remove(cacheKey);
                    }
                    catch { /* 忽略缓存清理错误 */ }
                    
                    // 记录错误并重新抛出异常，保留原始堆栈
                    KiviiContext.Logger.Error($"发票处理失败: {ex.Message}", ex);
                    throw; // 使用throw而不是throw ex，保留原始堆栈
                }
            }
            catch (Exception ex)
            {
                // 确保缓存被清理
                try {
                    cache.Remove(cacheKey);
                }
                catch { /* 忽略缓存清理错误 */ }
                
                // 确保数据库锁被释放
                if (lockAcquired)
                {
                    try {
                        conn.ExecuteScalar<int>("SELECT RELEASE_LOCK(@lockName)",
                            new { lockName = lockName });
                    }
                    catch { /* 忽略锁释放错误 */ }
                }
                
                // 记录错误日志并重新抛出异常
                KiviiContext.Logger.Error($"发票接收处理失败: {ex.Message}", ex);
                throw; // 使用throw而不是throw ex，保留原始堆栈
            }
        }
    }

    [Api(Description = "发票查询")]
    [RequiresAnyRole(MemberRoles.Everyone)]
    public class InvoiceQueryEx : RestfulExecution<Invoice>
    {
        #region QueryArgs
        public virtual int? Skip { get; set; }

        public virtual int? Take { get; set; }

        public virtual string OrderBy { get; set; }

        public string OrderByDesc { get; set; }

        public virtual string Include { get; set; }

        public virtual string Fields { get; set; }

        public string QueryKeys { get; set; }

        public string QueryValues { get; set; }

        #endregion

        public DateTime? BeginTime { get; set; }
        public DateTime? EndTime { get; set; }

        public bool? Debited { get; set; }//借票到账未到账查询

        public override object OnExecution(IRequest req, IResponse res)
        {
            (BeginTime != null && EndTime != null && BeginTime.Value > EndTime.Value).ThrowIfTrue("开始日期不能大于结束时间!");
            var conn = KiviiContext.GetOpenedDbConnection<Invoice>();
            var dynamicParams = Request.GetRequestParams();
            var autoQuery = Request.TryResolve<IAutoQueryDb>();
            autoQuery.IncludeTotal = true;
            var request = new RestfulQuery<Invoice>();
            request = request.PopulateWith(this);
            var sqlExpress = autoQuery.CreateQuery(Request, conn, request, dynamicParams);
            sqlExpress.Where(o => o.RootKvid == o.Kvid && o.Type != "Relation");
            //sqlExpress.Where(o => o.RootKvid == o.Kvid && o.OffsetKvid == Guid.Empty && o.Category == InvoiceApplyType.Debit.ToString());
            if (BeginTime != null) sqlExpress.And(o => o.OperateTime >= BeginTime.Value);
            if (EndTime != null) sqlExpress.And(o => o.OperateTime < EndTime.Value);
            if (Debited != null)
            {
                if (Debited.Value) sqlExpress.And(o => o.AmountPayment == o.Amount);
                else sqlExpress.And(o => o.AmountPayment < o.Amount);
            }
            var rtns = autoQuery.Execute(Request, conn, request, sqlExpress);

            if (Debited != null)
            {
                if (!Debited.Value)
                {
                    var queryAccount = conn.From<Account>();
                    queryAccount.Where(o => o.OrganizationKvid == KiviiContext.CurrentMember.OrganizationKvid && Sql.In(o.Type, AccountType.Balance, AccountType.Cash, AccountType.Pos));
                    queryAccount.Select(o => o.Kvid);

                    var queryPayment = conn.From<Payment>();
                    //queryPayment.Where(o => o.RootKvid == o.Kvid && o.OffsetKvid == Guid.Empty && o.AmountInvoice < o.Amount && (Sql.In(o.PayerAccountKvid, queryAccount)));
                    queryPayment.Where(o => o.Type == PaymentType.Split && o.OwnerKvid == KiviiContext.CurrentMember.DepartmentKvid && o.OffsetKvid == Guid.Empty && o.AmountInvoice < o.Amount && (Sql.In(o.PayerAccountKvid, queryAccount)));
                    queryPayment.Select(o => o.PayerName);
                    var paymentNames = conn.Column<string>(queryPayment).Distinct().ToList();
                    foreach (var item in rtns.Results)
                    {
                        if (paymentNames.Exists(o => o == item.PayerName))
                        {
                            item.Type = "Debit";
                        }
                    }
                    //rtns.Results = rtns.Results.OrderBy(o => o.Type).ToList();
                }
            }
            return rtns;
        }
    }

    [RequiresAnyRole(MemberRoles.Everyone)]
    public class InvoiceQuery : RestfulQuery<Invoice>
    { }

    [RequiresAnyRole(MemberRoles.Everyone)]
    public class InvoiceUpdate : RestfulUpdate<Invoice>
    {

    }

    [Api(Description = "发票更新")]
    [RequiresAnyRole(MemberRoles.Everyone)]
    public class InvoiceUpdateEx : RestfulExecution<Invoice>
    {
        public Invoice Item { get; set; }
        public override object OnExecution(IRequest req, IResponse res)
        {
            Item.ThrowIfNull("请传入要更新的信息!");
            Item.Kvid.ThrowIfEmpty("缺少更新Kvid!");
            Item.RemoveOnlyProperties(o => o.AmountPayment);
            Item.RemoveOnlyProperties(o => o.RootKvid);
            Item.RemoveOnlyProperties(o => o.AmountUntaxed);
            Item.RemoveOnlyProperties(o => o.AmountTax);
            var rtns = new RestfulUpdateResponse<Invoice>();
            rtns.Results = new List<Invoice>();
            if (Item.GetOnlyProperties().IsNullOrEmpty()) return rtns;
            var conn = KiviiContext.GetOpenedDbConnection<Invoice>();
            var exist = conn.SingleById<Invoice>(Item.Kvid);
            exist.ThrowIfNull("未找到要更新的发票信息!");
            if (exist.Type.ToLower() == "relation") throw new Exception("不支持的数据类型,无法更新!");
            if (Item.OffsetKvid != Guid.Empty) throw new Exception("无法更新已作废发票!");
            if (Item.SerialNumber == exist.SerialNumber) Item.RemoveOnlyProperties(o => o.SerialNumber);
            if (Item.OnlyPropertiesIsExist(o => o.SerialNumber))
            {
                var nodes = conn.Select<Invoice>(o => o.RootKvid == Item.Kvid && o.RootKvid != o.Kvid);
                if (!nodes.IsNullOrEmpty())
                {
                    foreach (var node in nodes)
                    {
                        node.SerialNumber = node.SerialNumber.Replace(exist.SerialNumber, Item.SerialNumber);
                        node.AddOnlyProperties(o => o.SerialNumber);
                        conn.UpdateOnly(node);
                    }
                }
            }
            if (Item.Amount == exist.Amount) Item.RemoveOnlyProperties(o => o.Amount);
            if (Item.OnlyPropertiesIsExist(o => o.Amount))
            {
                if (Item.Amount <= 0) throw new Exception("发票金额不能小于等于0!");
                var details = conn.Select<InvoiceDetail>(o => o.InvoiceKvid == Item.Kvid);
                if (exist.AmountPayment > 0) throw new Exception("无法更新发票金额!");
                if (exist.AmountUsed > 0) throw new Exception("无法更新发票金额!");
                Item.AmountUntaxed = Math.Round(Item.Amount / (1 + Item.TaxRate), 2);
                Item.AddOnlyProperties(o => o.AmountUntaxed);
                Item.AmountTax = Item.Amount - Item.AmountUntaxed;
                Item.AddOnlyProperties(o => o.AmountTax);

                var detail = new InvoiceDetail();
                if (!details.IsNullOrEmpty()) detail.PopulateInstance(details[0]);
                else detail.GoodsFullName = "检测费";
                detail.Kvid = Guid.NewGuid();
                detail.InvoiceKvid = exist.Kvid;
                detail.TaxRate = Item.TaxRate;
                detail.Amount = Item.Amount;
                detail.AmountUntaxed = Item.AmountUntaxed;
                detail.AmountTax = Item.AmountTax;
                detail.QuantityUnitPrice = Math.Round(detail.Amount / detail.Quantity, 2);
                detail.QuantityUnitPriceUntaxed = Math.Round(detail.AmountUntaxed / detail.Quantity, 2);
                detail.Remark = "手改修改发票金额,生成记录";

                var trans = conn.OpenTransaction();
                try
                {
                    conn.UpdateOnly(Item);
                    conn.Insert(detail);
                    var updateDetails = conn.From<InvoiceDetail>();
                    updateDetails = updateDetails.Update(o => o.Status);
                    updateDetails = updateDetails.Where(o => Sql.In(o.Kvid, details.ConvertAll(p => p.Kvid)));
                    conn.UpdateOnly<InvoiceDetail>(new InvoiceDetail { Status = -1 }, updateDetails);
                    trans.Commit();
                }
                catch (Exception ex)
                {
                    trans.Rollback();
                    throw ex;
                }
            }
            else if (Item.OnlyPropertiesIsExist(o => o.TaxRate))
            {
                (Item.TaxRate < 0 || Item.TaxRate > 1).ThrowIfTrue("发票税率取值不合法!");
                var details = conn.Select<InvoiceDetail>(o => o.InvoiceKvid == Item.Kvid);
                Item.AmountUntaxed = Math.Round(exist.Amount / (1 + Item.TaxRate), 2);
                Item.AddOnlyProperties(o => o.AmountUntaxed);
                Item.AmountTax = exist.Amount - Item.AmountUntaxed;
                Item.AddOnlyProperties(o => o.AmountTax);
                foreach (var detail in details)
                {
                    detail.TaxRate = Item.TaxRate;
                    detail.AddOnlyProperties(o => o.TaxRate);
                    detail.AmountUntaxed = Math.Round(detail.Amount / (1 + detail.TaxRate), 2);
                    detail.AddOnlyProperties(o => o.AmountUntaxed);
                    detail.AmountTax = detail.Amount - detail.AmountUntaxed;
                    detail.AddOnlyProperties(o => o.AmountTax);
                    detail.QuantityUnitPriceUntaxed = Math.Round(detail.AmountUntaxed / detail.Quantity, 2);
                    detail.AddOnlyProperties(o => o.QuantityUnitPriceUntaxed);

                }
                var trans = conn.OpenTransaction();
                try
                {
                    conn.UpdateOnly(Item);
                    foreach (var item in details)
                    {
                        conn.UpdateOnly(item);
                    }
                    trans.Commit();
                }
                catch (Exception ex)
                {
                    trans.Rollback();
                    throw ex;
                }

            }
            else conn.UpdateOnly(Item);
            rtns.Results.Add(Item);
            return rtns;
        }
    }

    [RequiresAnyRole(MemberRoles.Everyone)]
    public class InvoiceRead : RestfulRead<Invoice>
    {
        public bool IncludeDetail { get; set; } = true;

        public override bool OnPostRestfulRead(IRequest req, IResponse res, System.Data.IDbConnection dbConnection, IRestfulReadResponse<Invoice> rtns)
        {
            if (IncludeDetail)
            {
                var details = dbConnection.Select<InvoiceDetail>(o => o.InvoiceKvid == Kvid);
                rtns.Result.Details = new List<InvoiceDetail>();
                foreach (var item in details)
                {
                    var detail = new InvoiceDetail();
                    detail.PopulateWith(item);
                    rtns.Result.Details.Add(detail);
                }
            }
            return base.OnPostRestfulRead(req, res, dbConnection, rtns);
        }
    }
}
